#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Andrew Spangler
Licensed under GPLv3
This file is part of py_simple_morse_code
"""


# Tone for generated audio
DEFAULT_TONE = 1000  # Hz
# Min CW tone for live processing
DEFAULT_MIN_TONE = 600  # Hz
# Max CW tone for live processing
DEFAULT_MAX_TONE = 1500  # Hz
# Sample size for live processing
DEFAULT_SAMPLE_SIZE = 512
# Sample rate in samples / second
DEFAULT_SAMPLE_RATE = 32000
# High pass cutoff in Hz
DEFAULT_HIGH_PASS_FREQUENCY = 4
# Low pass cutoff in Hz
DEFAULT_LOW_PASS_FREQUENCY = 10000
# Default number of deadbeats to add to end of generated audio
DEFAULT_NUM_DEADBEATS = 0
# # Default duration of a beat in seconds
# DEFAULT_BEAT_DURATION = 0.1
# Default words per minute rate for transmit and receive
DEFAULT_WORDS_PER_MINUTE = 24
# Min time between allowed state changes
DEFAULT_STATE_DEBOUNCE = 0.01
# Durations for short and long pulses
DURATIONS = {
    ".": 1,  # Short pulse, one unit
    "-": 3,  # Long pulse, three units
}

UNKNOWN_CHAR = "?"

MORSE_CODE_LETTERS_MAP = {
    "a": ".-",
    "b": "-...",
    "c": "-.-.",
    "d": "-..",
    "e": ".",
    "f": "..-.",
    "g": "--.",
    "h": "....",
    "i": "..",
    "j": ".---",
    "k": "-.-",
    "l": ".-..",
    "m": "--",
    "n": "-.",
    "o": "---",
    "p": ".--.",
    "q": "--.-",
    "r": ".-.",
    "s": "...",
    "t": "-",
    "u": "..-",
    "v": "...-",
    "w": ".--",
    "x": "-..-",
    "y": "-.--",
    "z": "--..",
}
# Add uppercase
for char in list(MORSE_CODE_LETTERS_MAP.keys()):
    MORSE_CODE_LETTERS_MAP[char.upper()] = MORSE_CODE_LETTERS_MAP[char]

MORSE_CODE_NUMBERS_MAP = {
    "1": ".----",
    "2": "..---",
    "3": "...--",
    "4": "....-",
    "5": ".....",
    "6": "-....",
    "7": "--...",
    "8": "---..",
    "9": "----.",
    "0": "-----",
}

# Not 100% standard but widely used
MORSE_CODE_PUNCTUATION_MAP = {
    ".": ".-.-.-",
    ",": "--..--",
    ":": "---...",
    ";": "-.-.-.",
    "?": "..--..",
    "'": ".----.",
    "-": "-....-",
    "+": ".-.-.",
    "=": " -...-",
    "_": "..--.-",
    "/": "-..-.",
    "(": "-.--.",
    ")": "-.--.-",
    "&": ".-...",
    "$": "...-..-",
    "@": ".--.-.",
    '"': ".-..-.",
    "\b": "........",  # Backspace
}

# Make master char encoder map
CHAR_TO_MORSE_MAP = MORSE_CODE_LETTERS_MAP.copy()
CHAR_TO_MORSE_MAP.update(MORSE_CODE_NUMBERS_MAP)
CHAR_TO_MORSE_MAP.update(MORSE_CODE_PUNCTUATION_MAP)

# Invert char to morse lookup.
MORSE_TO_CHAR_MAP = {v: k for k, v in CHAR_TO_MORSE_MAP.items()}
# Add unknown char and newline handlers
MORSE_TO_CHAR_MAP.update({UNKNOWN_CHAR: UNKNOWN_CHAR, "\n": "\n"})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Andrew Spangler
Licensed under GPLv3
This file is part of py_simple_morse_code
"""


import time
import pyaudio
import numpy as np

# Set up audio
_MORSE_AUDIO_HOOK = pyaudio.PyAudio()

from .CONSTANTS import (
    DEFAULT_TONE,
    DEFAULT_NUM_DEADBEATS,
    DEFAULT_WORDS_PER_MINUTE,
    DEFAULT_STATE_DEBOUNCE,
    DEFAULT_SAMPLE_RATE,
    DURATIONS,
    UNKNOWN_CHAR,
    CHAR_TO_MORSE_MAP,
    MORSE_TO_CHAR_MAP,
)


def encode_string_to_morse(
    in_str: str,  # Your input string
    short_char: str = ".",  # Change to overwrite short pulse char
    long_char: str = "-",  # Change to overwrite long pulse char
    sep_char: str = " ",  # Change to overwrite separator char
    replace_on_unknown: str = UNKNOWN_CHAR,  # Unknown char, None to raise error
    verbose: bool = True,
) -> str:
    """Encodes text to morse code dots and dashes. *Returns a string*"""
    if verbose:
        print(f"Encoding string {in_str} to morse...")

    # Sanitize input string, runs until fully sanitized
    prev_string = None  # Used to check if string changed since last pass
    while not prev_string == in_str:
        prev_string = in_str
        in_str = in_str.strip().replace("\n", "").replace("  ", " ")

    morse = ""
    for c in in_str:
        if c == " ":
            # add another sep_char to make word separator
            morse += sep_char
            continue

        char = CHAR_TO_MORSE_MAP.get(c)
        if not char:
            if not replace_on_unknown:
                raise LookupError(
                    f"Could not find morse pattern for char {c} in lookup table"
                )
            char = CHAR_TO_MORSE_MAP.get(replace_on_unknown)
        morse += char + sep_char

    morse = morse.strip(sep_char).replace(".", short_char).replace("-", long_char)

    if verbose:
        print(f"Encoded string: {morse}\n")

    return morse


def decode_morse_to_string(
    morse: str,
    char_sep: str = " ",
    word_sep: str = "  ",
    replace_on_unknown: bool = UNKNOWN_CHAR,
) -> str:
    """Decodes morse dots and dashes to plaintext. *Returns a string*"""
    # Handle standard slash wordsep
    morse = morse.strip(" ").replace(" / ", "  ")

    out_str = ""

    for word in morse.split(word_sep):
        for char_pattern in word.split(char_sep):
            decoded_char = MORSE_TO_CHAR_MAP.get(char_pattern)
            if not decoded_char:
                if not replace_on_unknown:
                    raise LookupError(
                        f"Could not find char for morse pattern {char_pattern} in lookup table"
                    )
                decoded_char = replace_on_unknown
            out_str += decoded_char
        out_str += " "

    return out_str.lstrip(UNKNOWN_CHAR).lstrip("\n")


def encode_morse_to_beats(morse: str, verbose: bool = True) -> list:
    """Converts dots and dashes to a beats list. *Returns a list of bools*"""
    if not len(morse):
        return ""
    if verbose:
        print(f"Encoding morse {morse} to beats")
    beats, first = [], True
    for word in morse.split("  "):
        if not first:  # Skip wordsep on first
            # Expand charsep to wordsep
            beats.extend([False for i in range(4)])
        first = False
        for char_pattern in word.split(" "):
            for symbol in char_pattern:
                beats.extend(True for i in range(DURATIONS[symbol]))
                beats.append(False)  # Beat between chars
            # Expand beatsep to charsep
            beats.extend([False for i in range(2)])
    if verbose:
        print(f"Encoded morse: {beats}\n")
    return beats


def encode_string_to_beats(in_str: str, verbose: bool = True) -> list:
    """Converts a plaintext string to a beats list. *Returns a list of bools*"""
    morse = encode_string_to_morse(in_str, verbose=verbose)
    return encode_morse_to_beats(morse, verbose=verbose)


def make_morse_visual_from_beats(beats: list) -> str:
    """Converts a beats list to a visual representation in string form \
    using block chars (unicode char 2588). *Returns a string*"""
    visual = ""
    # chars = ["_", "-"]
    chars = [" ", "\u2588"]
    for b in beats:
        visual += chars[b]
    return visual


def encode_beats_to_waveform(
    beats: list,
    tone: int = DEFAULT_TONE,
    words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    verbose: bool = True,
    volume: float = 0.75,  # Prevent clipping
    deadbeats: int = DEFAULT_NUM_DEADBEATS,  # silent beets at the end of the waveform
) -> np.ndarray:
    """Encode a beats list into a waveform. *Returns a float32 1-dimensional numpy array*"""

    if verbose:
        print(f"Encoding beats to waveform: {beats}")

    start = time.time()

    beat_duration = calculate_beat_duration_from_wpm(words_per_minute)

    # Generate 3 beat sample at given tone
    tone_sample = (
        np.sin(
            2 * np.pi * np.arange(sample_rate * beat_duration * 3) * tone / sample_rate
        )
    ).astype(np.float32)

    silent_sample = np.zeros((len(tone_sample) // 3), dtype=np.float32)

    samples = {
        1: tone_sample[: len(tone_sample) // 3],
        3: tone_sample,
    }

    waveform = []
    # Walk through beats and assemble waveform from samples
    beatcount = len(beats)
    index, count, state = 0, None, None
    while index < beatcount:
        if state is None:
            state = beats[index]
            count = 1
            index += 1
        # Counts the number of beats between state changes
        while index < beatcount and beats[index] == state:
            count += 1
            index += 1

        if state:
            waveform.extend(samples[count])
        else:
            for i in range(count):
                waveform.extend(silent_sample)

        count = 0
        state = None

    # Add deadbeats to end
    for i in range(deadbeats):
        waveform.extend(silent_sample)

    waveform = np.asarray(waveform, dtype=np.float32)

    end = time.time()
    if verbose:
        print(f"Encoding took {end-start}")
        print(f"Waveform is {len(waveform)} samples\n")

    return waveform


def play_waveform(
    waveform: bytes, format=pyaudio.paFloat32, sample_rate=DEFAULT_SAMPLE_RATE
) -> None:
    """Plays a waveform. *Return None*"""
    stream = _MORSE_AUDIO_HOOK.open(
        format=format, channels=1, rate=sample_rate, output=True
    )
    stream.write(waveform.tobytes())
    stream.stop_stream()
    stream.close()


def play_string(
    in_str: str,
    tone: int = DEFAULT_TONE,
    words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    verbose: bool = True,
) -> None:
    """Converts plaintext to a waveform and play it on system speakers. *Returns None*"""
    beats = encode_string_to_beats(in_str, verbose)
    waveform = encode_beats_to_waveform(
        beats, tone, words_per_minute, sample_rate, verbose
    )
    play_waveform(waveform, sample_rate=DEFAULT_SAMPLE_RATE)


def play_morse(
    morse: str,
    tone: int = DEFAULT_TONE,
    words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    verbose: bool = True,
) -> None:
    """Converts dots and dashes to a waveform and plays it on system speakers. *Returns None*"""
    beats = encode_morse_to_beats(morse, verbose)
    waveform = encode_beats_to_waveform(
        beats, tone, words_per_minute, sample_rate, verbose
    )
    play_waveform(waveform, rate=sample_rate)


def encode_string_to_waveform(
    in_str: str,
    tone: int = DEFAULT_TONE,
    words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    verbose: bool = True,
) -> np.ndarray:
    """Encode a plaintext string to a waveform. *Returns a float32 1-dimensional numpy array*"""
    beats = encode_string_to_beats(in_str)
    return encode_beats_to_waveform(beats, tone, words_per_minute, sample_rate, verbose)


def encode_morse_to_waveform(
    morse: str,
    tone: int = DEFAULT_TONE,
    words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    verbose: bool = True,
) -> np.ndarray:
    """Encode a morse string to a waveform. *Returns a float32 1-dimensional numpy array*"""
    beats = encode_morse_to_beats(morse)
    return encode_beats_to_waveform(beats, tone, words_per_minute, sample_rate, verbose)


def calculate_beat_duration_from_wpm(words_per_minute: int) -> float:
    """Caclulate beat duration from words per minute. *Returns a float*"""
    return (1200.0 / words_per_minute) / 1000.0


class MorseCodeTranslator:
    """A low-level morse code translator. Inputs should be debounced / sanitized \
    before being passed to the .update(state:bool) method. Tolerance only affects \
    word-sep deadbeats to account for hesitation / early resumes between words."""

    def __init__(
        self,
        words_per_minute: int = DEFAULT_WORDS_PER_MINUTE,
        debounce_time: float = DEFAULT_STATE_DEBOUNCE,
        tolerance: int = 0,  # 1: low tolerance, 2: high tolerance
    ):
        self.start = time.time()

        self.words_per_minute = words_per_minute
        self.debounce_time = debounce_time
        self.tolerance = tolerance

        self.last_state = None
        self.last_state_change = None
        self.unparsed_content = ""
        self.parsed_content = ""

    def update(self, state: bool) -> None:
        """
        Call this whenever input state changes.
        Alternatively call this at a set frequency with the current state.
        *Returns None*
        """
        if state == self.last_state:
            return
        timestamp = time.time()
        last_state = self.last_state
        last_state_change = self.last_state_change
        self.last_state = state
        self.last_state_change = timestamp
        if not last_state_change:  # skip on first pulse
            return

        dt = timestamp - last_state_change
        if dt < self.debounce_time:
            # Ignore too fast state changes
            self.last_state_change = last_state_change
            return
        beat_count = round(dt / calculate_beat_duration_from_wpm(self.words_per_minute))
        # Handle large beat errors at wide tolerance
        if self.tolerance > 1 and beat_count in (5, 6, 8, 9):
            beat_count = 7
        elif self.tolerance > 0 and beat_count in (6, 8):
            beat_count = 7

        if beat_count:
            if not state:  # Going off high state
                if beat_count == 1:
                    self.unparsed_content += "."
                elif beat_count == 3:
                    self.unparsed_content += "-"
                else:
                    print(f"UNKNOWN BEAT COUNT - HIGH:{beat_count}")
                    self.unparsed_content += "?"
            else:  # Going into high state
                if beat_count == 1:
                    self.unparsed_content += ""
                elif beat_count == 3:
                    self.unparsed_content += " "
                elif beat_count == 7:
                    self.unparsed_content += "  "
                elif beat_count > 7:
                    # Add newline when pause is greater than wordsep
                    # Keep in mind that this may take up to 10 beats to trigger
                    # at higher tolerance values
                    self.unparsed_content += "\n"
                else:
                    print(f"UNKNOWN BEAT COUNT - LOW:{beat_count}")
        self.unparsed_content = self.unparsed_content.lstrip()
        self.parsed_content = decode_morse_to_string(self.unparsed_content)



from .py_simple_morse_code import (
    MorseCodeTranslator,
    encode_string_to_morse,
    decode_morse_to_string,
    encode_morse_to_beats,
    encode_string_to_beats,
    make_morse_visual_from_beats,
    encode_beats_to_waveform,
    play_waveform,
    play_string,
    play_morse,
    encode_string_to_waveform,
    encode_morse_to_waveform,
)
from .signal_processing import SignalProcessor
from .CONSTANTS import (
    DEFAULT_TONE,
    DEFAULT_MIN_TONE,
    DEFAULT_MAX_TONE,
    DEFAULT_SAMPLE_SIZE,
    DEFAULT_SAMPLE_RATE,
    DEFAULT_HIGH_PASS_FREQUENCY,
    DEFAULT_LOW_PASS_FREQUENCY,
    DEFAULT_NUM_DEADBEATS,
    DEFAULT_WORDS_PER_MINUTE,
    DEFAULT_STATE_DEBOUNCE,
    DURATIONS,
    UNKNOWN_CHAR,
    MORSE_CODE_LETTERS_MAP,
    MORSE_CODE_NUMBERS_MAP,
    MORSE_CODE_PUNCTUATION_MAP,
    CHAR_TO_MORSE_MAP,
    MORSE_TO_CHAR_MAP,
)
from .version import version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Andrew Spangler
Licensed under GPLv3
This file is part of py_simple_morse_code
"""


import pyaudio
import numpy as np
import scipy
import scipy.signal

from .CONSTANTS import (
    DEFAULT_MIN_TONE,
    DEFAULT_MAX_TONE,
    DEFAULT_SAMPLE_SIZE,
    DEFAULT_SAMPLE_RATE,
    DEFAULT_HIGH_PASS_FREQUENCY,
    DEFAULT_LOW_PASS_FREQUENCY,
)


class SignalProcessor:
    """Non-blocking CW signal processor. The .process() method returns true if a CW tone was found in the signal."""

    def __init__(
        self,
        mic_index: int = 0,
        min_tone: int = DEFAULT_MIN_TONE,
        max_tone: int = DEFAULT_MAX_TONE,
        sample_size: int = DEFAULT_SAMPLE_SIZE,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        high_pass_frequency: int = DEFAULT_HIGH_PASS_FREQUENCY,
        low_pass_frequency: int = DEFAULT_LOW_PASS_FREQUENCY,
    ):
        self.mic_index = mic_index
        self.sample_size = sample_size
        self.sample_rate = sample_rate
        self.min_tone = min_tone
        self.max_tone = max_tone
        # Calculate nyquist once
        self.nyquist = sample_rate / 2
        # Bandpass filter values to reduce noise
        self.high_pass_frequency = high_pass_frequency
        self.low_pass_frequency = low_pass_frequency
        # Set up audio session
        self.audio_session = pyaudio.PyAudio()
        self.stream = None

    def start_session(self) -> None:
        """Start the audio stream. *Returns None*"""
        if not self.stream is None:
            raise ValueError("Active signal processor session exists")
        # Open pyaudio stream
        self.stream = self.audio_session.open(
            format=pyaudio.paFloat32,
            channels=1,
            input_device_index=self.mic_index,
            rate=self.sample_rate,
            input=True,
            frames_per_buffer=self.sample_size,
        )

    def end_session(self) -> None:
        """Ends the audio stream. *Returns None*"""
        if self.stream is None:
            raise ValueError("No active signal processor session exists to stop")
        self.stream.stop_stream()
        self.stream.close()

    def process(self) -> bool:
        """Process a chunk of the audio stream from the buffer. *Returns None*"""
        if self.stream is None:
            raise ValueError("No active signal processor session exists")
        # This only blocks if there is an insufficient buffer
        buf = self.stream.read(self.sample_size)
        samples = np.frombuffer(buf, dtype=np.float32)
        # Apply fft
        fft_array = np.fft.fft(samples)
        # Calculate butterworth coefficients for bandpass
        denominator, numerator = scipy.signal.butter(
            5,  # Use butterworth order number of 5
            [
                self.high_pass_frequency / self.nyquist,
                self.low_pass_frequency / self.nyquist,
            ],
            btype="band",
        )
        # Apply bandpass filter to reduce signal noise
        fft_array = scipy.signal.lfilter(denominator, numerator, fft_array)
        # Get frequency bins
        freq_array = np.fft.fftfreq(samples.size)
        # Get the index of the peak frequency
        peak_index = np.argmax(np.abs(fft_array) ** 2)
        # Get peak frequency
        freq = np.abs(freq_array[peak_index]) * self.sample_rate
        # Return bool indicating peak frequency is or isn't in specified range
        return bool(int(freq) > self.min_tone and int(freq) < self.max_tone)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Andrew Spangler
Licensed under GPLv3
This file is part of py_simple_morse_code
Some portions of the GUI code found in this file come from py_simple_ttk
"""

import time
import platform
import threading
import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont
from tkinter.filedialog import asksaveasfilename
from typing import Callable
from scipy.io import wavfile
from .py_simple_morse_code import (
    decode_morse_to_string,
    encode_morse_to_beats,
    encode_string_to_morse,
    encode_string_to_waveform,
    make_morse_visual_from_beats,
    MorseCodeTranslator,
    play_string,
)

from .signal_processing import SignalProcessor

from .CONSTANTS import (
    DEFAULT_SAMPLE_RATE,
    DEFAULT_WORDS_PER_MINUTE,
)

if __name__ == "__main__":

    def play(in_str: str, tone: int, words_per_minute: int) -> None:
        threading.Thread(
            target=lambda: play_string(
                in_str,
                tone=tone,
                words_per_minute=words_per_minute,
                verbose=False,
            )
        ).start()

    def export(filename: str, in_str: str, tone: int, words_per_minute: float) -> None:
        if filename:
            waveform = encode_string_to_waveform(
                in_str, tone=tone, words_per_minute=words_per_minute
            )
            wavfile.write(filename, DEFAULT_SAMPLE_RATE, waveform)

    # Widgets with scroll bars that appear when needed and supporting code
    class Scroller(object):
        """Use to wrap an object with scrollbars"""

        def __init__(self, parent: ttk.Frame):
            try:
                vsb = ttk.Scrollbar(parent, orient="vertical", command=self.yview)
            except:
                pass
            hsb = ttk.Scrollbar(parent, orient="horizontal", command=self.xview)

            try:  # Fails if vsb instantiation failed
                self.configure(yscrollcommand=self._scroll(vsb))
            except:
                pass
            self.configure(xscrollcommand=self._scroll(hsb))

            self.grid(column=0, row=0, sticky="nsew")
            try:
                vsb.grid(column=1, row=0, sticky="ns")
            except:
                pass
            hsb.grid(column=0, row=1, sticky="ew")

            parent.grid_columnconfigure(0, weight=1)
            parent.grid_rowconfigure(0, weight=1)

            methods = (
                tk.Pack.__dict__.keys()
                | tk.Grid.__dict__.keys()
                | tk.Place.__dict__.keys()
            )

            for method in methods:
                if method[0] != "_" and method not in ("config", "configure"):
                    setattr(self, method, getattr(parent, method))

        @staticmethod
        def _scroll(bar) -> Callable:
            """Hide and show scrollbar as needed."""

            def hide(start, end) -> None:
                start, end = float(start), float(end)
                if start <= 0.0 and end >= 1.0:
                    bar.grid_remove()
                else:
                    bar.grid()
                bar.set(start, end)

            return hide

        def __str__(self) -> str:
            return str(self.parent)

    def _create_container(func: Callable) -> Callable:
        """Creates a tk Frame with a given parent, and uses this new frame to place the scrollbars and the widget."""

        def wrapped(cls, parent, **kw) -> object:
            container = ttk.Frame(parent)
            container.bind("<Enter>", lambda e: _bound_to_mousewheel(e, container))
            container.bind("<Leave>", lambda e: _unbound_to_mousewheel(e, container))
            return func(cls, container, **kw)

        return wrapped

    def _bound_to_mousewheel(event, widget) -> None:
        child = widget.winfo_children()[0]
        if platform.system() == "Windows" or platform.system() == "Darwin":
            child.bind_all("<MouseWheel>", lambda e: _on_mousewheel(e, child))
            child.bind_all("<Shift-MouseWheel>", lambda e: _on_shiftmouse(e, child))
        else:
            child.bind_all("<Button-4>", lambda e: _on_mousewheel(e, child))
            child.bind_all("<Button-5>", lambda e: _on_mousewheel(e, child))
            child.bind_all("<Shift-Button-4>", lambda e: _on_shiftmouse(e, child))
            child.bind_all("<Shift-Button-5>", lambda e: _on_shiftmouse(e, child))

    def _unbound_to_mousewheel(event, widget) -> None:
        if platform.system() == "Windows" or platform.system() == "Darwin":
            widget.unbind_all("<MouseWheel>")
            widget.unbind_all("<Shift-MouseWheel>")
        else:
            widget.unbind_all("<Button-4>")
            widget.unbind_all("<Button-5>")
            widget.unbind_all("<Shift-Button-4>")
            widget.unbind_all("<Shift-Button-5>")

    def _on_mousewheel(event, widget) -> None:
        if platform.system() == "Windows":
            widget.yview_scroll(-1 * int(event.delta / 120), "units")
        elif platform.system() == "Darwin":
            widget.yview_scroll(-1 * int(event.delta), "units")
        else:
            if event.num == 4:
                widget.yview_scroll(-1, "units")
            elif event.num == 5:
                widget.yview_scroll(1, "units")

    def _on_shiftmouse(event, widget) -> None:
        if platform.system() == "Windows":
            widget.xview_scroll(-1 * int(event.delta / 120), "units")
        elif platform.system() == "Darwin":
            widget.xview_scroll(-1 * int(event.delta), "units")
        else:
            if event.num == 4:
                widget.xview_scroll(-1, "units")
            elif event.num == 5:
                widget.xview_scroll(1, "units")

    class ScrolledText(Scroller, tk.Text):
        """Scrolled Text with SuperWidget mixin"""

        __desc__ = """Scrolled Text SuperWidget"""

        @_create_container
        def __init__(self, parent, widgetargs: dict = {}, **kw):
            tk.Text.__init__(
                self,
                parent,
                wrap=tk.WORD,
                **kw,
            )
            # Some systems don't bind this automatically
            self.bind("<Control-Key-a>", self.select_all, "+")
            Scroller.__init__(self, parent)
            # Create widget proxy
            self._orig = self._w + "_orig"
            self.tk.call("rename", self._w, self._orig)
            self.tk.createcommand(self._w, self._event_generate)

        def _event_generate(self, tk_call: str, *args) -> tk.Event:
            """Proxy method to generate an event when text contents change"""
            try:
                result = self.tk.call((self._orig, tk_call) + args)
            except Exception:
                return None
            if tk_call in ("insert", "delete", "replace"):
                self.event_generate("<<Modified>>")
            return result

        def select_all(self, event=None) -> None:
            """Selects all text. `Returns None`"""
            self.tag_add(tk.SEL, "1.0", tk.END)
            self.mark_set(tk.INSERT, "1.0")
            self.see(tk.INSERT)

        def set(self, val: str) -> None:
            """Sets the text. `Returns a String`"""
            state = self["state"]
            self.configure(state=tk.NORMAL)
            self.clear()
            self.insert("1.0", val)
            self.configure(state=state)

        def get(self, start: str = "1.0", end: str = tk.END):
            """Returns the contents of the text box with optional start/end kwargs. `Returns a String`"""
            return tk.Text.get(self, start, end)

        def clear(self) -> None:
            """Empties the text box. `Returns None`"""
            self.delete("1.0", tk.END)

        def get_cursor(self):
            """Get the current location of the cursor. `Returns None`"""
            return tk.Text.index(self, tk.INSERT)

        def set_cursor(self, col: int, row: int) -> None:
            """Sets the cursor to a given col / row. `Returns None`"""
            text_widget_name.mark_set(tk.INSERT, "%d.%d" % (col, row))

        def enable(self) -> None:
            """Enable Text box"""
            self.config(state="normal")

        def disable(self) -> None:
            self.config(state="disable")

    class CopyBox(ttk.Frame):
        """Scrolled Text with "Copy to Clipboard" Button"""

        __desc__ = "A widget with a scrolled textbox and button that copies the textbox contents to the user's clipboard. Useful for form output, etc."

        def __init__(self, parent: ttk.Frame, **kw):
            ttk.Frame.__init__(self, parent)
            self.button = ttk.Button(
                self, text="Copy to Clipboard", command=self._on_click
            )
            self.button.pack(side=tk.BOTTOM, fill="x", expand=False, pady=(0, 5))
            self.text = ScrolledText(self, **kw)
            self.text.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        def _on_click(self) -> None:
            self.button.configure(text="Copied!")
            self.clipboard_clear()
            self.clipboard_append(self.get())
            self.after(1000, lambda: self.button.configure(text="Copy To Clipboard"))

        def enable(self) -> None:
            """Enable CopyBox"""
            self.text.enable()

        def disable(self) -> None:
            """Disable CopyBox"""
            self.text.disable()

        def get(self) -> None:
            """Get CopyBox contents"""
            return self.text.get()

        def set(self, val: str) -> None:
            """Set CopyBox Contents"""
            self.text.set(val)

        def clear(self) -> None:
            """Clear CopyBox Contents"""
            self.text.clear()

    class TextToMorseTab(ttk.Frame):
        def __init__(self, parent: ttk.Frame):
            tk.Frame.__init__(self, parent)

            input_frame = ttk.LabelFrame(
                self, text="Text Input", style="Bold.TLabelframe"
            )
            input_frame.pack(fill="x", side="top", anchor="n", padx=10, pady=10)

            self.entry = ScrolledText(input_frame, height=2)
            self.entry.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=10
            )
            self.entry.bind("<<Modified>>", self.on_change, add="+")

            output_frame = ttk.LabelFrame(
                self, text="Outputs", style="Bold.TLabelframe"
            )
            output_frame.pack(
                side="top", fill="x", anchor="n", pady=(0, 10), padx=10, expand=True
            )

            #############
            out_frame = ttk.LabelFrame(output_frame, text="Morse")
            out_frame.pack(fill="x", expand=True, side="top", anchor="n", padx=10)
            self.output = CopyBox(out_frame, height=2)
            self.output.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=(0, 10)
            )
            self.output.text.disable()

            #############
            bits_frame = ttk.LabelFrame(output_frame, text="Bitstream Output")
            bits_frame.pack(fill="x", expand=True, side="top", anchor="n", padx=10)
            self.beats = CopyBox(bits_frame, height=2)
            self.beats.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=(0, 10)
            )
            self.beats.text.disable()

            #############
            visual_frame = ttk.LabelFrame(output_frame, text="Visual Output")
            visual_frame.pack(fill="x", expand=True, side="top", anchor="n", padx=10)
            self.visual = CopyBox(visual_frame, height=2)
            self.visual.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=(0, 10)
            )
            self.visual.text.disable()

            #############
            audio_frame = ttk.LabelFrame(output_frame, text="Audio Output")
            audio_frame.pack(fill="x", side="top", anchor="n", padx=10, pady=(0, 10))

            options_frame = ttk.Frame(audio_frame)
            options_frame.pack(fill="x", padx=10, pady=10)

            tone_frame = ttk.Frame(options_frame)
            tone_frame.pack(fill="x", padx=10, side="left")
            ttk.Label(tone_frame, text="Tone (Hz)").pack(fill="x", side="left")
            self.tone_var = tk.StringVar(value="800")
            ttk.Combobox(
                tone_frame,
                values=[i for i in range(800, 1500, 100)],
                textvariable=self.tone_var,
            ).pack(fill="x", expand=True, side="right")

            wpm_frame = ttk.Frame(options_frame)
            wpm_frame.pack(fill="x", side="right")
            ttk.Label(wpm_frame, text="Words Per Minute").pack(fill="x", side="left")
            self.duration_var = tk.StringVar(value="24")
            ttk.Combobox(
                wpm_frame,
                values=list(range(1, 31)),
                textvariable=self.duration_var,
            ).pack(fill="x", expand=True, padx=10, side="right")

            ttk.Button(audio_frame, text="Play", command=self.on_play).pack(
                fill="x", expand=False, anchor="n", padx=10, pady=(0, 10)
            )

            ttk.Button(audio_frame, text="Export", command=self.on_export).pack(
                fill="x", expand=False, anchor="n", padx=10, pady=(0, 10)
            )

        def on_change(self, event):
            morse = encode_string_to_morse(self.entry.get(), verbose=False)
            self.output.text.enable()
            self.output.set(morse.replace("  ", " / "))
            self.output.text.disable()

            beats = encode_morse_to_beats(morse, verbose=False)
            self.beats.text.enable()
            self.beats.set("".join([["0", "1"][b] for b in beats]))
            self.beats.text.disable()

            visual = make_morse_visual_from_beats(beats)
            self.visual.text.enable()
            self.visual.set(visual)
            self.visual.text.disable()

        def on_play(self) -> None:
            play(
                self.entry.get(),
                int(self.tone_var.get()),
                float(self.duration_var.get()),
            )

        def on_export(self) -> None:
            filetypes = [("Wave File", "*.wav")]
            filename = asksaveasfilename(
                filetypes=filetypes, defaultextension=filetypes
            )

            export(
                filename,
                self.entry.get(),
                int(self.tone_var.get()),
                float(self.duration_var.get()),
            )

    class MorseToTextTab(ttk.Frame):
        def __init__(self, parent: ttk.Frame):
            tk.Frame.__init__(self, parent)

            input_frame = ttk.LabelFrame(
                self, text="Morse Input", style="Bold.TLabelframe"
            )
            input_frame.pack(fill="x", side="top", anchor="n", padx=10, pady=10)
            self.entry = ScrolledText(input_frame, height=2)
            self.entry.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=10
            )
            self.entry.bind("<<Modified>>", self.on_change, add="+")

            output_frame = ttk.LabelFrame(self, text="Output", style="Bold.TLabelframe")
            output_frame.pack(fill="x", side="top", anchor="n", pady=(0, 10), padx=10)
            self.output = CopyBox(output_frame, height=2)
            self.output.pack(
                fill="both", expand=True, side="top", anchor="n", padx=10, pady=(0, 10)
            )
            self.output.text.disable()

        def on_change(self, event):
            text = decode_morse_to_string(self.entry.get())
            self.output.set(text)

    class LiveDecoderTab(ttk.Frame):
        def __init__(self, parent: ttk.Frame):
            tk.Frame.__init__(self, parent)

            self.signal_processor = SignalProcessor()
            self.decoder = MorseCodeTranslator(tolerance=2, words_per_minute=24)

            self.enabled = False
            self.enable_button_text_var = tk.StringVar(value="Enable")
            tk.Button(
                self, textvariable=self.enable_button_text_var, command=self.toggle
            ).pack(fill="x", expand=False, padx=10, pady=10)

            options_frame = ttk.Frame(self)
            options_frame.pack(fill="x", padx=10, pady=(0, 10))

            tone_frame = ttk.Frame(options_frame)
            tone_frame.pack(fill="x", padx=10, side="left")
            ttk.Label(tone_frame, text="Approximate Tone (Hz)").pack(
                fill="x", side="left"
            )
            self.tone_var = tk.StringVar(value="1000")
            tone_combobox = ttk.Combobox(
                tone_frame,
                values=[i for i in range(800, 1500, 100)],
                textvariable=self.tone_var,
            )
            tone_combobox.pack(fill="x", expand=True, side="right")
            tone_combobox.bind("<<ComboboxSelected>>", self.on_tone_change)

            wpm = ttk.Frame(options_frame)
            wpm.pack(fill="x", side="right")
            ttk.Label(wpm, text="Words Per Minute").pack(fill="x", side="left")
            self.wpm_var = tk.StringVar(value="24")
            wpm_combobox = ttk.Combobox(
                wpm,
                values=list(range(5, 31)),
                textvariable=self.wpm_var,
            )
            wpm_combobox.pack(fill="x", expand=True, padx=10, side="right")
            wpm_combobox.bind("<<ComboboxSelected>>", self.on_wpm_change)

            unparsed_frame = ttk.LabelFrame(self, text="Raw Input")
            unparsed_frame.pack(fill="both", expand=True, padx=10)
            self.unparsed_var = tk.StringVar()
            tk.Label(unparsed_frame, textvariable=self.unparsed_var).pack(
                fill="both", expand=True, side="top", anchor="center"
            )

            parsed_frame = ttk.LabelFrame(self, text="Decoded")
            parsed_frame.pack(fill="both", expand=True, padx=10)
            self.parsed_var = tk.StringVar()
            self.parsed_label = tk.Label(parsed_frame, textvariable=self.parsed_var)
            self.parsed_label.pack(
                fill="both", expand=True, side="top", anchor="center"
            )
            tk.Button(self, text="Reset", command=self.clear).pack(
                fill="x", expand=False, padx=10, pady=10
            )

            # Post setup
            self.on_tone_change()
            self.on_wpm_change()

        def toggle(self) -> None:
            self.enabled = not self.enabled
            self.enable_button_text_var.set(["Enable", "Disable"][self.enabled])

        def clear(self) -> None:
            self.decoder.unparsed_content = ""
            self.decoder.parsed_content = ""
            self.unparsed_var.set(self.decoder.unparsed_content)
            self.parsed_var.set(self.decoder.parsed_content)

        def update(self) -> None:
            self.decoder.update(self.signal_processor.process())
            self.unparsed_var.set(self.decoder.unparsed_content)
            self.parsed_var.set(self.decoder.parsed_content)

        def on_wpm_change(self, event=None) -> None:
            wpm = int(self.wpm_var.get())
            print(f"Decoder wpm set to {wpm}")
            self.decoder.words_per_minute = wpm

        def on_tone_change(self, event=None) -> None:
            tone = int(self.tone_var.get())
            print(f"Live processor signal tone set to {tone}")
            self.signal_processor.min_tone = tone - 400
            self.signal_processor.max_tone = tone + 400

    class UpdateThread(threading.Thread):
        def __init__(self, gui, *args, **kw):
            self.gui = gui
            threading.Thread.__init__(self, *args, **kw)

        def run(self) -> None:
            """Thread loop for keep gui updated"""
            try:
                self.gui.live_decoder_tab.signal_processor.start_session()  # open audio resources
                while True:
                    if self.gui.live_decoder_tab.enabled:
                        self.gui.live_decoder_tab.update()
                    else:
                        time.sleep(0.1)
            except KeyboardInterrupt:
                self.gui.live_decoder_tab.signal_processor.end_session()

    class GUI(tk.Tk):
        def __init__(self):
            tk.Tk.__init__(self)
            self.title("Morse GUI")
            fnt = tkFont.nametofont("TkDefaultFont").actual()
            self.bold_font = (fnt["family"], fnt["size"], "bold")
            self.style = ttk.Style()
            self.style.theme_use("winnative")
            self.style.configure("Bold.TLabelframe.Label", font=self.bold_font)

            # Add all the tabs
            notebook = ttk.Notebook(self)
            notebook.pack(fill="x")
            notebook.add(TextToMorseTab(self), text="Text->Morse")
            notebook.add(MorseToTextTab(self), text="Morse->Text")
            self.live_decoder_tab = LiveDecoderTab(self)  # Needed by update thread
            notebook.add(self.live_decoder_tab, text="Live Decoder")

    gui = GUI()
    UpdateThread(gui).start()
    gui.mainloop()
